! Array based string code

System_file;

Property arstr;

! Objects wanting to use ArStrings should provide the property array arstr, which must be at least 4 bytes long

Constant MAX_ARSTR 128;

Class ArString(MAX_ARSTR)
  with data 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0, ! 32 bytes of string
       next 0; ! Next ArString in the list

[ PrintArStr obj len pos;
  len = obj.&arstr->0;
  if (len < obj.#arstr)
  {
    for (pos=1:pos<=len:pos++)
      PrintCharSafe(obj.&arstr->pos);
    return;
  }
  ! Else string object(s)
  obj = obj.&arstr-->1;
  while (len > 0)
  {
    PrintCharSafe(obj.&data->pos);
    pos++;
    if (pos == obj.#data)
    {
      pos = 0;
      obj = obj.next;
    }
    len--;
  }
];

[ ArStrNull s tmp; ! Nullify s
  if (s.&arstr->0 < s.#arstr)
  {
    s.&arstr->0 = 0;
    return;
  }
  s.&arstr->0 = 0;
  while (s.&arstr-->1 ~= nothing)
  {
    tmp = s.&arstr-->1;
    s.&arstr-->1 = tmp.next;
    ArString.destroy(tmp);
  }
];

[ ArStrLen str;
  return str.&arstr->0;
];

[ ArStrGetChar str pos; ! Return char at pos 'pos'
  if (pos >= str.&arstr->0)
    return 0;
  if (str.&arstr->0 < str.#arstr)
    return str.&arstr->(pos+1);
  str = str.&arstr-->1;
  while ((str ~= nothing) && (pos >= str.#data))
  {
    pos = pos - str.#data;
    str = str.next;
  }
  if (str == nothing)
    return 0; ! Shouldn't happen
  return str.&data->pos;
];

[ ArStrSetChar str pos val tmp tmp2;
!  print "ArStrSetChar: ", pos, " to ", val, ", curlen ", str.&arstr->0, ", shortlen ", str.#arstr, "^";
  if (str.&arstr->0 < str.#arstr) ! Currently a short string
  {
    if (pos+1 < str.#arstr) ! Trying to write into short string?
    {
      str.&arstr->(pos+1) = val;
      if (str.&arstr->0 < pos+1)
        str.&arstr->0 = pos+1; ! Increase length
      return;
    }
    ! Else need to create a long string
    tmp = ArString.create();
    for (tmp2=0:tmp2<str.&arstr->0:tmp2++)
      tmp.&data->tmp2 = str.&arstr->(tmp2+1);
    str.&arstr-->1 = tmp;
  }
  tmp = str.&arstr-->1;
  ! Else writing to a long string
  while (pos > tmp.#data)
  {
    if (tmp.next == nothing)
      tmp.next = ArString.create();
    tmp = tmp.next;
    pos = pos - tmp.#data;
  }
  tmp.&data->pos = val;
  if (str.&arstr->0 < pos+1)
    str.&arstr->0 = pos+1;
];

[ ArStrTrim str len tmp tmp2; ! Trim string to length len; can also be used to expand string
  if (str.&arstr->0 < len)
  {
    ArStrSetChar(str,len-1,0); ! Make longer
    return;
  }
  if (str.&arstr->0 < str.#arstr) ! Short string?
  {
    str.&arstr->0 = len;
    return;
  }
  ! Else long string
  if (len < str.#arstr)
  {
    ! Transform into short
    while ((str.&arstr-->1).next) ! Get rid of all but last ArString
    {
      tmp = (str.&arstr-->1).next;
      (str.&arstr-->1).next = tmp.next;
      ArString.destroy(tmp);
    }
    tmp2 = str.&arstr-->1;
    for (tmp=0:tmp<len:tmp++)
      str.&arstr->(tmp+1) = tmp2.&data->tmp;
    ArString.destroy(tmp2);
    str.&arstr->0 = len;
    return;
  }
  ! Keep as long string
  str.&arstr->0 = len;
  tmp = str.&arstr-->1;
  while (len > tmp.#data)
  {
    tmp = tmp.next;
    len = len - tmp.#data;
  }
  while (tmp.next ~= nothing) ! Should be none left after this one
  {
    tmp2 = tmp.next;
    tmp.next = tmp2.next;
    ArString.destroy(tmp2);
  }
];

[ ArStrCpy from to tmp; ! Copy (i.e. set) string from from to to
  ArStrTrim(to,from.&arstr->0);
  for (tmp=0:tmp<from.&arstr->0:tmp++)
    ArStrSetChar(to,tmp,ArStrGetChar(from,tmp));
];

[ ArStrEq a b tmp; ! Returns true if strings are equal
  if (a.&arstr->0 ~= b.&arstr->0)
    rfalse;
  for (tmp=0:tmp<a.&arstr->0:tmp++)
    if (ArStrGetChar(a,tmp) ~= ArStrGetChar(b,tmp))
      rfalse;
  rtrue;
];

[ ArStrEqToByteArray ar by tmp;
  if (ar.&arstr->0 ~= by->0)
    rfalse;
  for (tmp=0:tmp<by->0:tmp++)
    if (ArStrGetChar(ar,tmp) ~= by->(tmp+1))
      rfalse;
  rtrue;
];

[ ArStrToByteArray from to tmp; ! Assumes 'to' is long enough
  for (tmp=0:tmp < from.&arstr->0:tmp++)
    to->(tmp+1) = ArStrGetChar(from,tmp);
  to->0 = from.&arstr->0;
];

[ ArStrFromByteArray from to tmp;
  ArStrTrim(to,from->0);
  for (tmp=0:tmp<from->0:tmp++)
    ArStrSetChar(to,tmp,from->(tmp+1));
];

[ ArStrOneEq o a; ! Is o a 1-char string of 'a'?
  if (o == nothing)
    rfalse;
  if ((o.&arstr->0) ~= 1)
    rfalse;
  if ((o.&arstr->1) ~= a)
    rfalse;
  rtrue;
];

[ ArStrTwoEq o a b; ! Is o a 2-char string of 'ab'?
  if (o == nothing)
    rfalse;
  if ((o.&arstr->0) ~= 2)
    rfalse;
  if ((o.&arstr->1) ~= a)
    rfalse;
  if ((o.&arstr->2) ~= b)
    rfalse;
  rtrue;
];

[ ArStrThreeEq o a b c; ! Is o a 3-char string of 'abc'?
  if (o == nothing)
    rfalse;
  if ((o.&arstr->0) ~= 3)
    rfalse;
  if ((o.&arstr->1) ~= a)
    rfalse;
  if ((o.&arstr->2) ~= b)
    rfalse;
  if ((o.&arstr->3) ~= c)
    rfalse;
  rtrue;
];

[ PrintCharSafe c; ! Print char c, or a printable version of it
  if ((c == 10) || (c == 13))
    print "^";
  else if ((c < 32) || (c == 127) || (c > 255))
    print "?"; ! Should really do it in bold or something
  else
    print (char) c;
];
